/* 
 * File:   Letter_1.c
 * Author: a
 *
 * Created on 25. May 2018, 07:45
 */

#include <stdio.h>
#include <stdlib.h>
#include <xc.h>
#include "Init.h"


// CONFIG1
#pragma config FOSC = INTOSC    // Oscillator Selection (INTOSC oscillator: I/O function on CLKIN pin)
#pragma config WDTE = OFF       // Watchdog Timer Enable (WDT disabled)
#pragma config PWRTE = ON       // Power-up Timer Enable (PWRT enabled)
#pragma config MCLRE = ON       // MCLR Pin Function Select (MCLR/VPP pin function is MCLR)
#pragma config CP = OFF         // Flash Program Memory Code Protection (Program memory code protection is disabled)
#pragma config CPD = OFF        // Data Memory Code Protection (Data memory code protection is disabled)
#pragma config BOREN = OFF      // Brown-out Reset Enable (Brown-out Reset disabled)
#pragma config CLKOUTEN = OFF   // Clock Out Enable (CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin)
#pragma config IESO = OFF       // Internal/External Switchover (Internal/External Switchover mode is disabled)
#pragma config FCMEN = OFF      // Fail-Safe Clock Monitor Enable (Fail-Safe Clock Monitor is disabled)

// CONFIG2
#pragma config WRT = ALL        // Flash Memory Self-Write Protection (000h to FFFh write protected, no addresses may be modified by EECON control)
#pragma config PLLEN = ON      // PLL Enable (4x PLL enabled) fr 32 MHz
#pragma config STVREN = OFF     // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will not cause a Reset)
#pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), low trip point selected.)
#pragma config LVP = OFF        // Low-Voltage Programming Enable (High-voltage on MCLR/VPP must be used for programming)
/*
 * 
 */
#define _XTAL_FREQ  32000000 //needed for __delay...
#define Status RA6 //Status LED blinks: PIC is working fine, connected to RA6 pin
#define CKP SSP1CON1bits.CKP //to make the programm compatible to other PICs
#define D_nA SSP1STATbits.D_nA //the same
#define R_nW SSP1STATbits.R_nW //the same
#define SSPBUF SSP1BUF //and so on
#define SSPIF SSP1IF //...
#define Strobe RB7 //put the value of the stored data to the output pins of the 4094
#define Data RB6 //input pin of the 4094
#define Clock RA7 //clock pin of the 4094


char weiss[8]; //array for the "white" values
char schwarz[8]; //array for the "black" values
/**************************************************************/
char Schwellwert [8]; //threshold values
/**************************************************************/
char A_D_Count; //Counter for the A_D_value
char A_D_Wert[8]; //array for the detected AD values
char Letter_BIT[8]; //AD-value of each sensor
char Channel [] = {0,1,2,3,4,7,9,10}; //Analogchannels from the pins: RA0..4;RB5,3,2
int Tick; //Counter for TMR0-overflows
char junk; //transfered I2C-Adress
char Letter_Wert; //Bitpattern produced by phototransistors
char Master_1; //first value sent by master
char Master_2; //second value sent by master

char Bitmuster; //Bitpattern of the sensor
char Encoder_count; //amount of engine-turns, about 6 per cm, 2.5" wheel.
char Buchstabe; //ASCII value of the detectet letter
char H = 'H'; //ASCII-Code of letter H
char S = 'S'; //ASCII-Code of letter S
char U = 'U'; //ASCII-Code of letter U
char LED_Count; //which LED is currently on?

bit Master_ruft; //Flag: Master calls Slave
bit Flag_8_Werte; //All 8 values are determined

bit I2C_Read; //Master wants to read
bit I2C_Write; //No writing at the time
char I2C_Count; //counter for switch instruction

char z; //can be useful for tests

void init (void); //initialize chip
void Bitmuster_erzeugen (void); //generate bit pattern 
void Bitmuster_auswerten(void); //interpret bit pattern
void LEDs_clear (void); //all LEDs off, Safetyfeature!!! DON'T remove.
void Next_LED (void); //turn on next LED
void kalibrieren(void); //calibrate letter-sensor
void interrupt Alle (void); //All possible interrupts

/*main programm*/
void main(void) {
    init(); //initialize Chip 
    LEDs_clear(); //Important safetyfeature to be sure that all LED's are turned off
//Slave adress of the letter sensors, one is in 0x15, the other in 0x16
    
/**************************************************************************/
    SSP1ADD = 21<<1; //slave address = 21, gets shifted to the left by one
/**************************************************************************/
    
    for (char i = 0; i < 8; i++) { //Deleting the black and white values, probably not necessary...
        A_D_Wert[i] = 0; 
        weiss[i] = 0; 
        schwarz[i] = 0; 
    }
//    Bitmuster = 123; //for tests
    while(1) { //main loop
    //calibrate
        if ((Master_1 == 255) && (Master_2 == 255)) { //robot slowly drives past the line or was driven
            kalibrieren(); //get min-max-values for black and white
            Letter_Wert = A_D_Wert[3]; //Letter value is send to the master
        }
    //wait for signal to start, set threshold for recognizing either black or white
        if (Master_1 == 255 && Master_2 == 254) { //Roboter stands still
            Bitmuster_erzeugen(); //create bit pattern
            Letter_Wert = Bitmuster; //Letter value is send to the master
            for (char i = 0; i < 8; i++) { //determine threshold values
                Schwellwert[i] = weiss[i] -((weiss[i] - schwarz[i]) >> 1); //average of black and white -> threshold
            }
        }
    //now it starts: in this example only the ASCII code of the letter "H" is
    //send to the Master. You have to send from the Master a value between
    //40 and 170. That might be the value the IR-Sensors are given, if there is
    //a wall. You have to change Bitmuster_auswerten() to get other values than "H"
        if (Master_1 <= 170) { //If the robot has the right distance to the wall...
            if (Master_1 >= 40) { 
                Bitmuster_erzeugen(); //(generate bitpattern)
                Letter_Wert = Buchstabe; //Letter value is send to the master
                //either 0 = no letter. or ASCII-value of H, S or U, for example
                //switch - case allows you to change the threshold value
            }
        }
    }
}

// calibrate sensors  
//let the robot slowly drive past the line!
//(values are written into the EEPROM) Not implemented yet!!!
//stop the robot 
void kalibrieren(void) {
    for (char i = 0; i < 8; i++) { //evaluate 8 AD-values 
        if (A_D_Wert[i] > weiss[i]) { //if the current value is bigger than the previous one...
            weiss[i] = A_D_Wert[i]; //set new max-value
        }
        if (A_D_Wert[i] < schwarz[i]) { //if the current value is smaller than the previous one...
            schwarz[i] = A_D_Wert[i]; //set new min-value
        }
    }
}

//delete all LEDs
//LED1 (Eagle Board, in line with the PIC) is indeed LED1
//SAFETYFEATURE!!!! Otherwise eyes could be hurt
void LEDs_clear (void) {
    Data = 0; //serial data is NULL
    for (char i = 0; i < 8; i++) { //eight "Bits"
        Clock = 1; //clock into register
        Clock = 0;
    }
    Strobe = 1; //send to latch
    Strobe = 0; //...
}

//turn on next LED
//!!!!!!!!!DON'T EVER CHANGE THIS FUNCTION!!!!!!!!!!!
//The LED are so bright that they can hurt your eyes, if they're constantly on!!
//...and you wouldn't notice until it's to late!!!!!
void Next_LED (void) { //Funtion is called in an interrupt
        //push a 1 through the register of the 4094
        if (LED_Count == 0) {
            Data = 1; //At first a one...
        }
        else {
            Data = 0; //...then push zeros
        }
        Clock = 1; //right now...
        Clock = 0; //...
        Strobe = 1; //put it into the Latch...
        Strobe = 0; //...the LED appropriated to the "1" is luminated
        LED_Count++; //Next LED
        if (LED_Count == 8) { //8 LEDs
            LED_Count = 0; //Wieder von vorne
/***********Dont know why 6? But it works*************************************/
            A_D_Count = 6; //synchronize counts
        }
}

//create bitpattern
//is only executed if all 8 values are determined. Bitmuster-auswerten function included
void Bitmuster_erzeugen (void) {
    if (Flag_8_Werte) { //are all 8 values determined? (see Interrupt)
        Flag_8_Werte = 0; //delete flag, for a new beginning 
        Bitmuster = 0; //delete old bitpattern
        for (char i = 0; i < 8; i++) { //create new bitpattern out of the A_D_values
            Bitmuster = Bitmuster << 1; //the first time, a 0 is shifted!
            if (Letter_BIT[i] < Schwellwert[i]) { //black = 1, white = 0
                Bitmuster = (Bitmuster | 0b00000001); //bit 0 is set to 1
            }
        }
        Bitmuster_auswerten(); //Not implemented yet, only a "H" is generated
    }
}

//This is the important part. You have to fill!
void Bitmuster_auswerten(void) {
/****Test*****************************************************************/
    Buchstabe = 'H';
/*************************************************************************/
}

//Interrupts
void interrupt Alle (void) {
    //Status = 1; //can be useful for debugging (see below)
    //Prescaler 1:4, 256 * 4 = 1024 cycles / (32[MHz]/4) = 128us
    //about 1ms is needed to get all 8 values
    //At a speed of 0.4 m/s the robot drives about 0.4mm
//TMR0 overflow interrupt
    if (TMR0IF) { //has the TMR0 register turned from 255 to 0
        TMR0IF = 0; //No more TMR0-interrupt, otherwise the interrupt will never end
        Tick++; //Makes an LED flash to indicate that the PIC is working properly
            if (Tick >= 2300) {
                Status = !Status; //~ 200ms
                Tick = 0; //start again
            }
        //At least 128us have been past so we can be sure, that the AD-value has changed
        //AN 0, 1, 2, 3, 4, 7, 9, 10 (s. Channel)
        //The LED [A_D_Count + 1] is on, the A_D_value [A_D_Cont] is determined
        A_D_Wert [A_D_Count] = ADRESH; //the previously determined AD value is saved
        Letter_BIT[A_D_Count] = A_D_Wert[A_D_Count]; //for working with the values
        A_D_Count++; //next channel
        if (A_D_Count == 8) {
            A_D_Count = 0; //eight values
            Flag_8_Werte = 1; //All 8 values are determined
        }
        //The LED [A_D_Count] has been on for 120us, A_D_value [A_D_Count] is getting initialized
        ADCON0 = 0b10000001 | (Channel[A_D_Count] << 2); //Bit 6-2 determine the Channel
/*******Now we have to wait at least 5us = 40 cycles, before A_D_ coversion can be started*/
        __delay_us(5); //
//start AD-conversion
        //5us should be over by now
        //Analogvalue is re-initialized every 0,128ms
        GO_nDONE = 1; //start AD-conversion (Bit 2 von ADCON0, S. 127)
        //light up next LED
        Next_LED();
        //conversion takes 40us, TMR0-interrupt updates the value every 128us
        //Status = 0; //see above
    }

//Interrupt on RB0, encoder of the motor
    if (INTF) { //interrupt by changing edge to RB0?
            INTF = 0; //delete flag
            Encoder_count++; //increment encoder
            OPTION_REG = OPTION_REG ^ INTEDG; //toggle, falling/rising edge
    }
    
//I2C Interrupt, I2C service routine
    if (SSPIF){//does SSPBUF have a value?
        SSPIF = 0; //delete flag
            if(!D_nA) { //Was an address transferred?
                junk = SSPBUF;//read buffer, dummy read to clear the BF bit
                if (R_nW) { //Master wants to read, LSB of the Adressbyte = 0
                    I2C_Read = 1; //Master wants to read
                    I2C_Write = 0; //No writing at the time
                    I2C_Count = 0; //step-counter
                    //After a read-request, the byte thats about to be transferred is send to the USB-bus
                    //without an SSP-Interrupt in between
/*******************example "Letter_Wert"******************************************/
                    SSPBUF = Letter_Wert; //write desired value into buffer
/**********************************************************************************/ 
                    CKP = 1; //Release Clock Stretch
                //After transfer, the slave keeps the CLK-channel on GND (LO)
                //release clock stretch, Master can now write again
                    Master_ruft = 1; //indicates that a reading-/writingprocess is completed, might be unnecessary
                }
                else { //Master wants to write, LSB of the adressbyte = 1
                    I2C_Read = 0; //Master doesn't want to reaad
                    I2C_Write = 1; //Master is currently writing
                    I2C_Count = 0; //step-counter
                    CKP = 1; //Release Clock
                }
            }
            //second and perhaps third interrupt
            else { //D_nA = 1, some data has been transferred
                if(I2C_Write) { //Master wants to write
                    switch(I2C_Count) {
                        case 0: //first Interrupt after adress + writing
                            Master_1 = SSPBUF; //safe transferred byte, deletes BF 
                            I2C_Count++; //Interruptcounter
                            CKP = 1; //set bus free
                            break; //leave function switch()
                        case 1: //second interrupt after adress + writing 
                            Master_2 = SSPBUF; //safe transferred byte
                            I2C_Count++; //Interruptcounter
                            I2C_Write = 0; //writing is ended
                            CKP = 1; //set bus free
                            Master_ruft = 1; //indicates that a reading-/writingprocess is completed, might be unnecessary
                            break; //leave function switch()
                        default : //this point should never be reached
                            break; //leave function switch()
                    }
                }
                if(I2C_Read) { //master wants to read
                    //in case multiple things should be transferred at once,
                    //this is, where the routine would come
                    //will probably never be used
                    CKP = 1; //set bus free
                    I2C_Read = 0;
                }
            }
    }
}


